import React, {} from "react";
import { Link as RouterLink } from "react-router-dom";
import { Document } from "../accessor/document";

import { Box, Button, Grid, Link, Paper, Theme, Typography, useTheme } from "@mui/material";
import { TagChip } from "../component/tagchip";
import { ThumbnailContainer } from "../page/reader/reader";

import DocumentAccessor from "../accessor/document";

export const makeContentInfoUrl = (id: number) => `/doc/${id}`;
export const makeContentReaderUrl = (id: number) => `/doc/${id}/reader`;

const useStyles = (theme: Theme) => ({
	thumbnail_content: {
		maxHeight: "400px",
		maxWidth: "min(400px, 100vw)",
	},
	tag_list: {
		display: "flex",
		justifyContent: "flex-start",
		flexWrap: "wrap",
		overflowY: "hidden",
		"& > *": {
			margin: theme.spacing(0.5),
		},
	},
	title: {
		marginLeft: theme.spacing(1),
	},
	infoContainer: {
		padding: theme.spacing(2),
	},
	subinfoContainer: {
		display: "grid",
		gridTemplateColumns: "100px auto",
		overflowY: "hidden",
		alignItems: "baseline",
	},
	short_subinfoContainer: {
		[theme.breakpoints.down("md")]: {
			display: "none",
		},
	},
	short_root: {
		overflowY: "hidden",
		display: "flex",
		flexDirection: "column",
		[theme.breakpoints.up("sm")]: {
			height: 200,
			flexDirection: "row",
		},
	},
	short_thumbnail_anchor: {
		background: "#272733",
		display: "flex",
		alignItems: "center",
		justifyContent: "center",
		[theme.breakpoints.up("sm")]: {
			width: theme.spacing(25),
			height: theme.spacing(25),
			flexShrink: 0,
		},
	},
	short_thumbnail_content: {
		maxWidth: "100%",
		maxHeight: "100%",
	},
});

export const ContentInfo = (props: {
	document: Document;
	children?: React.ReactNode;
	classes?: {
		root?: string;
		thumbnail_anchor?: string;
		thumbnail_content?: string;
		tag_list?: string;
		title?: string;
		infoContainer?: string;
		subinfoContainer?: string;
	};
	gallery?: string;
	short?: boolean;
}) => {
	const theme = useTheme();
	const document = props.document;
	const url = props.gallery === undefined ? makeContentReaderUrl(document.id) : makeContentInfoUrl(document.id);
	return (
		<Paper
			sx={{
				display: "flex",
				height: props.short ? "400px" : "auto",
				overflow: "hidden",
				[theme.breakpoints.down("sm")]: {
					flexDirection: "column",
					alignItems: "center",
					height: "auto",
				},
			}}
			elevation={4}
		>
			<Link
				component={RouterLink}
				to={{
					pathname: makeContentReaderUrl(document.id),
				}}
			>
				{document.deleted_at === null ? (
					<ThumbnailContainer content={document} />
				) : (
					<Typography variant="h4">Deleted</Typography>
				)}
			</Link>
			<Box>
				<Link variant="h5" color="inherit" component={RouterLink} to={{ pathname: url }}>
					{document.title}
				</Link>
				<Box>
					{props.short ? (
						<Box>
							{document.tags.map((x) => (
								<TagChip key={x} label={x} clickable tagname={x} size="small"></TagChip>
							))}
						</Box>
					) : (
						<ComicDetailTag
							tags={document.tags}
							path={document.basepath + "/" + document.filename}
							createdAt={document.created_at}
							deletedAt={document.deleted_at != null ? document.deleted_at : undefined}
						/>
					)}
				</Box>
				{document.deleted_at != null && (
					<Button
						onClick={() => {
							documentDelete(document.id);
						}}
					>
						Delete
					</Button>
				)}
			</Box>
		</Paper>
	);
};
async function documentDelete(id: number) {
	const t = await DocumentAccessor.del(id);
	if (t) {
		alert("document deleted!");
	} else {
		alert("document already deleted.");
	}
}

function ComicDetailTag(prop: {
	tags: string[] /*classes:{
    tag_list:string
}*/;
	path?: string;
	createdAt?: number;
	deletedAt?: number;
}) {
	let allTag = prop.tags;
	const tagKind = ["artist", "group", "series", "type", "character"];
	let tagTable: { [kind: string]: string[] } = {};
	for (const kind of tagKind) {
		const tags = allTag.filter((x) => x.startsWith(kind + ":")).map((x) => x.slice(kind.length + 1));
		tagTable[kind] = tags;
		allTag = allTag.filter((x) => !x.startsWith(kind + ":"));
	}
	return (
		<Grid container>
			{tagKind.map((key) => (
				<React.Fragment key={key}>
					<Grid item xs={3}>
						<Typography variant="subtitle1">{key}</Typography>
					</Grid>
					<Grid item xs={9}>
						<Box>
							{tagTable[key].length !== 0
								? tagTable[key].map((elem, i) => {
										return (
											<>
												<Link to={`/search?allow_tag=${key}:${encodeURIComponent(elem)}`} component={RouterLink}>
													{elem}
												</Link>
												{i < tagTable[key].length - 1 ? "," : ""}
											</>
										);
									})
								: "N/A"}
						</Box>
					</Grid>
				</React.Fragment>
			))}
			{prop.path != undefined && (
				<>
					<Grid item xs={3}>
						<Typography variant="subtitle1">Path</Typography>
					</Grid>
					<Grid item xs={9}>
						<Box>{prop.path}</Box>
					</Grid>
				</>
			)}
			{prop.createdAt != undefined && (
				<>
					<Grid item xs={3}>
						<Typography variant="subtitle1">CreatedAt</Typography>
					</Grid>
					<Grid item xs={9}>
						<Box>{new Date(prop.createdAt).toUTCString()}</Box>
					</Grid>
				</>
			)}
			{prop.deletedAt != undefined && (
				<>
					<Grid item xs={3}>
						<Typography variant="subtitle1">DeletedAt</Typography>
					</Grid>
					<Grid item xs={9}>
						<Box>{new Date(prop.deletedAt).toUTCString()}</Box>
					</Grid>
				</>
			)}
			<Grid item xs={3}>
				<Typography variant="subtitle1">Tags</Typography>
			</Grid>
			<Grid item xs={9}>
				{allTag.map((x) => (
					<TagChip key={x} label={x} clickable tagname={x} size="small"></TagChip>
				))}
			</Grid>
		</Grid>
	);
}
