import { type Context, DefaultContext, DefaultState, Next } from "koa";
import Router from "koa-router";
import { createReadableStreamFromZip, entriesByNaturalOrder, readZip } from "../util/zipwrap";
import type { ContentContext } from "./context";
import { since_last_modified } from "./util";
import type { ZipReader } from "@zip.js/zip.js";
import type { FileHandle } from "node:fs/promises";
import { Readable, Writable } from "node:stream";

/**
 * zip stream cache.
 */
const ZipStreamCache: {
	[path: string]: [{
		reader: ZipReader<FileHandle>,
		handle: FileHandle
	}, number]
} = {};

async function acquireZip(path: string) {
	if (!(path in ZipStreamCache)) {
		const obj = await readZip(path);
		ZipStreamCache[path] = [obj, 1];
		// console.log(`acquire ${path} 1`);
		return obj.reader;
	}
	const [ret, refCount] = ZipStreamCache[path];
	ZipStreamCache[path] = [ret, refCount + 1];
	// console.log(`acquire ${path} ${refCount + 1}`);
	return ret.reader;
}

function releaseZip(path: string) {
	const obj = ZipStreamCache[path];
	if (obj === undefined) throw new Error("error! key invalid");
	const [ref, refCount] = obj;
	// console.log(`release ${path} : ${refCount}`);
	if (refCount === 1) {
		const { reader, handle } = ref;
		reader.close().then(() => {
			handle.close();
		});
		delete ZipStreamCache[path];
	} else {
		ZipStreamCache[path] = [ref, refCount - 1];
	}
}

async function renderZipImage(ctx: Context, path: string, page: number) {
	const image_ext = ["gif", "png", "jpeg", "bmp", "webp", "jpg"];
	// console.log(`opened ${page}`);
	const zip = await acquireZip(path);
	const entries = (await entriesByNaturalOrder(zip)).filter((x) => {
		const ext = x.filename.split(".").pop();
		return ext !== undefined && image_ext.includes(ext);
	});
	if (0 <= page && page < entries.length) {
		const entry = entries[page];
		const last_modified = entry.lastModDate;
		if (since_last_modified(ctx, last_modified)) {
			return;
		}
		const read_stream = await createReadableStreamFromZip(zip, entry);
		const nodeReadableStream = new Readable();
		nodeReadableStream._read = () => { };

		read_stream.pipeTo(new WritableStream({
			write(chunk) {
				nodeReadableStream.push(chunk);
			},
			close() {
				nodeReadableStream.push(null);
				setTimeout(() => {
					releaseZip(path);
				}, 100);
			},
		}));

		ctx.body = nodeReadableStream;
		ctx.response.length = entry.uncompressedSize;
		// console.log(`${entry.name}'s ${page}:${entry.size}`);
		ctx.response.type = entry.filename.split(".").pop() as string;
		ctx.status = 200;
		ctx.set("Date", new Date().toUTCString());
		ctx.set("Last-Modified", last_modified.toUTCString());
	} else {
		ctx.status = 404;
	}
}

export class ComicRouter extends Router<ContentContext> {
	constructor() {
		super();
		this.get("/", async (ctx, next) => {
			await renderZipImage(ctx, ctx.state.location.path, 0);
		});
		this.get("/:page(\\d+)", async (ctx, next) => {
			const page = Number.parseInt(ctx.params.page);
			await renderZipImage(ctx, ctx.state.location.path, page);
		});
		this.get("/thumbnail", async (ctx, next) => {
			await renderZipImage(ctx, ctx.state.location.path, 0);
		});
	}
}

export default ComicRouter;
